functor SeqSpanningTree(Seq : SEQUENCE) =
struct
   structure Rand = Random210
   type vertex = int

   fun isHead (rand : Rand.rand) (v : vertex) = (Rand.hashInt rand v) > 0

   fun spanningTree(E, n) =
   let
     open Seq

     fun ST(L : (vertex * int) seq, 
            E : ((vertex * vertex) * int) seq, 
            rand : Rand.rand) =
       if (length E = 0) then empty()
       else let
           (* hook edges are the ones that go from tail to head  *) 
           fun isHook((u,v),_) = not(isHead rand u) andalso (isHead rand v)
           val Hooks = filter isHook E

           (* each hook edge writes its head and label into its tail *)
           val L'  = inject (map (fn ((t,h), l) => (t,(h,l))) Hooks) L

           (* Identify edges that "won", these are added to the spanning tree *)
           val STE = filter (fn ((u,v),l) => #2(nth L' u) = l) Hooks

           (* Relabel edges to point to star center *)
           val E'  = map (fn ((u,v),l) => ((#1(nth L' u), #1(nth L' v)), l)) E

           (* Filter out self edges *)
           val E'' = filter (fn ((u,v),_) => u <> v) E'

           (* Make recursive call *)
           val STR = ST(L', E'', Rand.next rand)
         in
           (* append spanning tree edges to recursive call edges *)
           append(STE, STR)
         end

     (* tag each edge with its position *)
     val TE = tabulate (fn i => (nth E i, i)) (length E)

     (* array used for hooking *)
     val L = tabulate (fn i => (i,0)) n

     val T = ST(L, TE, Rand.fromInt 0)
   in
     (* for each label return the original edge *)
     map (fn (_,l) => nth E l) T
   end
end

