structure SuffixArray =
struct
  structure Seq = ArraySequence
  open Seq

  fun intPairCmp ((a0,a1),(b0,b1)) =
    case Int.compare(a0,b0)
     of EQUAL => Int.compare (a1,b1)
      | r => r

  fun zip(s,t) =  map2 (fn e => e) s t

  fun naiveRank cmp s =
  let (* compare only the payload, ignoring the index *)
      fun icmp((_,x),(_,y)) = cmp(x,y)
      fun isNeq(a,b) = if icmp(a,b)=EQUAL then 0 else 1
      val n = length s

      val s' = sort icmp (tabulate (fn i => (i,nth s i)) n)
      val incr = tabulate (fn i=>isNeq(nth s' i, nth s' (i+1))) (n-1)
      val (ranks,f) = scan op+ 0 incr
  in map (fn ((i,_),r)=>(i,r)) (zip(s',append(ranks,singleton f)))
  end


  fun suffixArray (S : char seq) =
  let val n = length S
      val zeros = tabulate (fn i => 0) n
      fun rank (sz:int) : (int*int) seq =
        case sz
          of 0 => raise Fail "shouldn't reach here"
           | 1 => naiveRank Char.compare S

           | _ => let val half = sz div 2
                      val rank' = inject (rank half) zeros
                      fun find j =
                        if j >= n then ~1 else nth rank' j
                      val s = tabulate
                        (fn i => (find i, find (i+half))) n

                  in naiveRank intPairCmp s
                  end

      val ppt = 1 + (IntInf.log2 (IntInf.fromInt  n))
  in map #1 (rank ppt)
  end
end
